#   Jabber.py: Implements the Jabber protocol, on top of XMLStream
#   Copyright (C) 1999  Daniel Arbuckle
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#    
#    Daniel Arbuckle can be contacted by email at djarb@wvc-omak.ctc.edu

import XMLStream.XMLInStream
import XMLStream.XMLOutStream
import XMLStream.Element

import socket
import string
import time

# There are only two requirements on packets: they must implement a
# constructor which accepts an Element.Element as it's only parameter, and
# they must implement an element_tree() method returning Element.Element
from JabberPackets import *

NoPackets = "There are no available packets."
BadSend = "Cannot send this type of packet"

try: # If we have threading, we want to take advantage of it.
    from threading import *
    from Queue import *
except: # If not, we still want our code to work.
    class Thread:
        "Fake thread class for use in nonthreaded environments"
        def __init__(self, group=None, target=None,
                     name=None, args=(), kwargs={}):
            return None

        def start(self):
            return None

        def setDaemon(self, daemonic):
            return None

    class Lock:
        "Fake lock class for use in nonthreaded environments"
        def acquire(self, blocking = 1):
            return None

        def release(self):
            return None
        
    class Queue:
        "Fake queue class for use in nonthreaded environments"
        def __init__(self, maxsize = 0):
            self.__data = []

        def qsize(self):
            return len(self.__data)

        def empty(self):
            return len(self.__data) == 0

        def put(self, item, block = 1):
            self.__data.append(item)

        def get(self, block = 1):
            if len(self.__data):
                raise NoPackets
            ret = self.__data[0]
            self.__data = self.__data[1:]
            return ret

JABBER_TAG = '<stream:stream to="%s" xmlns="jabber:client" xmlns:stream="http://etherx.jabber.org/streams">'

class UnrecognisedPacket:
    "Represents a recieved packet for which no wrapper is registered"
    def __init__(self, arg):
        if isinstance(arg, Element):
            self.msg = "Unrecognised packet type: " + arg.type
            self.tree = arg
        else:
            self.msg = arg
            self.tree = None

    def element_tree(self):
        raise BadSend, "UnrecognisedPacket"
    
class PacketQueue:
    """The application level interface to the Jabber stream
    protocol. The primary interface to PacketQueue objects is via
    their get() and put() methods. put() puts a packet onto the wire,
    and get() gets the next packet."""
    
    def __init__(self, server, port = 5222):
        self.__callback = None
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        addr = string.split(server, ":")
        try:
            sock.connect((addr[0], string.atoi(addr[1])))
        except IndexError:
            sock.connect((addr[0], port))

        self.__packets = {}
        self.__avail = Queue(0)        
        self.__update_lock = Lock()
        self.__stream_lock = Lock()
        
        auto_register_jabber_packets(self)
        
        self.__instream = XMLStream.XMLInStream.XMLInStream(sock, 1)
        self.__outstream = XMLStream.XMLOutStream.XMLOutStream(sock, 0)

        self.__instream.use_default_handler(self.__incoming_packet_handler)
        
        sock.send(JABBER_TAG % addr[0])

        self.thread = Thread(target = self.__update_periodically)
        self.thread.setDaemon(1)
        self.thread.start()

    def __update_periodically(self):
        while 1:
            self.update()
            time.sleep(0.1)

    def __incoming_packet_handler(self, tree):
        if not self.__packets.has_key(tree.type):
            self.__avail.put(UnrecognisedPacket(tree))
        else: self.__avail.put(self.__packets[tree.type](tree))
        if callable(self.__callback): self.__callback()
        
    def update(self):
        """Called to read available data from the line. If you are
        operating in a nonthreaded environment, it is your
        responsibility to call this method periodically."""
        self.__update_lock.acquire()
        self.__instream.update()
        self.__update_lock.release()

    def put(self, next):
        """Called to put a packet onto the wire. The single parameter
        is an object implementing the packet interface, specifically
        the element_tree() method"""
        self.__stream_lock.acquire()
        self.__outstream.put_next(next.element_tree())
        self.__stream_lock.release()

    def get(self):
        """Called to get the next packet from the incoming queue. If
        you are operating in a threaded environment, this will block
        until a packet is available. In a nonthreaded environment, a
        Jabber.NoPackets exception is thrown in the event of an empty
        queue."""
        return self.__avail.get() #Queue handles its own synchronization

    def available(self):
        "Returns the number of Packets in the incoming queue."
        return self.__avail.qsize()

    def register_packet_type(self, name, packet):
        """Called to add a new packet type to the list of understood
        packets. First parameter is a string of the format
        'namespace tag' which identifies the namespace and tag which
        the packet type handles. Second parameter will be called to
        construct a new packet whenever that tag is found."""
        
        self.__update_lock.acquire()
        self.__packets[name] = packet
        self.__update_lock.release()

    def set_callback(self, cb):
        """Allows you to set a callback which will be triggered on the 
        reception of a new packet. In a multithreaded environment,
        there are better ways of dealing with the problem of
        notification."""
        self.__callback = cb

    def set_debug(self, debug):
        """If debug != None, data sent and recieved on the wire will
        also be sent through debug. debug must implement the socket
        interface."""
        self.__stream_lock.acquire()
        self.__outstream.debug = debug
        self.__instream.debug = debug
        self.__stream_lock.release()
