#   XMLInStream: Transform an XML document into a tree of Element objects.
#   Copyright (C) 1999  Daniel Arbuckle
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#    
#    Daniel Arbuckle can be contacted by email at djarb@wvc-omak.ctc.edu

import xmllib
import string
import sys
import socket
import time

from Element import Element

class XMLInError:
    def __init__(self, msg = ""):
	self.msg = msg

    def __str__(self):
	return self.msg

class NoHandlerError(XMLInError):
    pass

class BadParameterError(XMLInError):
    pass

class XMLInStream(xmllib.XMLParser):
    def __init__(self, sock, depth = 0):
        xmllib.XMLParser.__init__(self)
	self.debug = None
        self.__depth = depth #Below this number of nestings, behave differently
        self.__sock = sock #Socket or Socketalike to read from
        self.__sock.setblocking(0) #Blocking is bad for event driven code
        self.__cdata = [] #LIFO of character data (NOT really XML cdata)
        self.__stack = [] #LIFO of incomplete elements
        self.__handlers = {} #Map of tags to callable objects
        self.__default_handler = None
        
    #Used to set a callback for a specific tag
    def use_handler(self, name, handler):
        if not callable(handler):
            raise BadParameterError("Handler must be callable.")
        self.__handlers[name] = handler

    #Used to set the default callback
    def use_default_handler(self, handler):
        if not callable(handler):
            raise BadParameterError("Handler must be callable.")
        self.__default_handler = handler

    #Triggers a handler based on the tag type
    def call_handler(self, element):
        if self.__handlers.has_key(element.type):
            self.__handlers[element.type](element)
        elif self.__default_handler:
            self.__default_handler(element)
        else:
            raise NoHandlerError("No handler for <%s> tags and no default handler.")
    
    #Reads data from the socket and feeds them to the parser. Should
    #be called frequently.
    def update(self):
        try:
            a = self.__sock.recv(100) #Could probably bump that up a bit :)
            if self.debug: self.debug.send(a)
            self.feed(a) 
        except socket.error:
            pass # Couldn't read data, so continue
        
    #A loop that repeatedly calls self.update(). Useful when you're just
    #going to put the instream in a seperate thread and let it go.
    def run(self):
        while 1:
            self.update()
            time.sleep(0.1)
    
    def handle_data(self, data):
        #This method takes care of data between <tag> and </tag> style tags.
        if len(self.__cdata) > 0:
            s = self.__cdata[-1]
            self.__cdata[-1] = s + data
    
    def unknown_starttag(self, tag, attr):
        t = Element(tag)
        t.update(attr) #The key/value pairs of attr are absobed into t
        self.__stack.append(t)
        self.__cdata.append("")

        #This if clause transforms elements all lower levels than __depth
        #into empty elements (except for attributes) at level __depth.
        #In most normal uses, that will be a no op. __depth is set greater
        #than 0 when the data are being read from a source which provides
        #them slowly, or interactively; i.e. the data must be processed
        #as they are recieved rather than waiting for the entire top level
        #element to be delivered. These "too low" tags aren't completely
        #ignored because their attributes generally contain important info.
        if len(self.__stack) <= self.__depth:
            self.call_handler(t)

    def unknown_endtag(self, tag):
        t = self.__stack.pop()
        d = self.__cdata.pop()
        t.cdata = d
        if len(self.__stack) == self.__depth: #It's at the right level to be
            self.call_handler(t)              #considered a complete fragment.
        elif len(self.__stack) > self.__depth: #It's nested too far down to be
            self.__stack[-1].add_child(t)      #self contained. It's a child.
            
    def syntax_error(self, error):
        #These are recoverable. Since I want this parser to be
        #tolerant, they are ignored.
        pass
